// -*- C++ -*-
// $RCSfile: petsc_solver.h,v $
// $Revision: 1.18 $
// $Author: redwing $
// $Date: 2004/02/19 21:28:09 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#ifdef HAVE_PETSC

#include <oofconfig.h>

#ifndef PETSCSOLVER_H
#define PETSCSOLVER_H
#include "engine/solver.h"
#include "petscvec.h"
#include "petscmat.h"
#include "petscsles.h"
#include "petsc_preconditioner.h"
#include "engine/csrmatrix.h"
#include "engine/matvec.h"
//#include "common/IO/threadstate.h"

#include <string>
int create_matrix();
//Encapsulation of the PETSc subroutines

class PETScEnvironment; /* Invokes PETSc universe. Universe is created if 
			   any solver is called. PETSc Universe must persist
			   until the end of OOF sesion. Otherwise, expect a
			   segmentation fault. PETSc universe is destroyed
			   when OOF exits.
			*/
//Solver classes
class PETScSolver; 
class PETScGMRESSolver;
class PETScBiCGSolver;

//message handlers
class PETScSolverStatus;
/*
  PETScSolverStatus extracts convergence/divergence information from 
  KSPGetConvergedReason() after returning from a solver callback. 
  Translates PETSc's state to something that is, hopefully, useful 
  to a human being.
*/


class PETScEnvironment {
private:
  int number_of_processors;
public:
  PETScEnvironment();
  //PETScEnvironment(int, char **);
  virtual ~PETScEnvironment();
};


class PETScSolver : public Solver {
public:
  PreconditionerWrap *pc;
  int max_iterations;
  double tolerance;
  int number_of_iterations;
  double residual_norm;
  double arlb;
  bool status_;
  /*
    Linear system will converge when  after the k-th iteration
    the residual of the k-th solution is smaller than the 
    largest of the tolerance or the absolute residual_lower_bound
  */
  double arub;
  /*
    The linear solver, whichever it is, will bail out if the relative 
    residual is larger than the absolute_residual_upper_bound
  */

  PETScSolver();//should create or check the environment
  PETScSolver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc);
  int get_max_iterations() const { return max_iterations; }
  double get_tolerance() const { return tolerance; }
  double get_arlb() const {return arlb; }
  double get_arub() const {return arub; }
  PreconditionerWrap *get_pc() const { return pc; }
  virtual void precompute(const CSRmatrix &A);
  virtual void postcompute();
  virtual ~PETScSolver(); //should remove all the objects
  virtual bool solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x);
  /*
    this has a non-void
    implementation because
    ALL the PETSc solvers have
    the same call.
    Internally, CSRMatrix and VECTOR_D objects are transformed into Mat
    and Vec, correspondingly.
  */
  //should probably encapsulate the Mat class
  virtual void setsolver(KSP &) const = 0; //derived class should set right solver
  virtual bool status() const { return status_;};
  virtual std::string message() const {return "PETSc solver";};
};

// Preconditioned gmres solver.

class PETScGMRESSolver : public PETScSolver {
public:

  PETScGMRESSolver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc)
    : PETScSolver::PETScSolver( max_iterations, tolerance, arlb, arub, pc)
  {}

  virtual void setsolver(KSP &) const;
  virtual std::string message() const;
};

//Preconditioned BiCG solver.

class PETScBiCGSolver : public PETScSolver {
public:
  PETScBiCGSolver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc)
    : PETScSolver::PETScSolver(max_iterations, tolerance, arlb, arub, pc)
  {}

  virtual void setsolver(KSP &) const;
  virtual std::string message() const;
};

//Preconditioned BiCG solver.

class PETScRichardsonSolver : public PETScSolver {
public:
  PETScRichardsonSolver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc)
    : PETScSolver::PETScSolver(max_iterations, tolerance, arlb, arub, pc)
  {}

  virtual void setsolver(KSP &) const;
  virtual std::string message() const;
};

class PETScChebychevSolver : public PETScSolver {
public:
  PETScChebychevSolver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc)
    : PETScSolver::PETScSolver(max_iterations, tolerance, arlb, arub, pc)
  {}

  virtual void setsolver(KSP &) const;
  virtual std::string message() const;
};


class PETScCGSolver : public PETScSolver {
public:
  PETScCGSolver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc)
    : PETScSolver::PETScSolver(max_iterations, tolerance, arlb, arub, pc)
  {}

  virtual void setsolver(KSP &) const;
  virtual std::string message() const;
};


class PETScBiCGStabSolver : public PETScSolver {
public:
  PETScBiCGStabSolver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc)
    : PETScSolver::PETScSolver(max_iterations, tolerance, arlb, arub, pc)
  {}

  virtual void setsolver(KSP &) const;
  virtual std::string message() const;
};

class PETScCGSquaredSolver : public PETScSolver {
public:
  PETScCGSquaredSolver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc)
    : PETScSolver::PETScSolver(max_iterations, tolerance, arlb, arub, pc)
  {}

  virtual void setsolver(KSP &) const;
  virtual std::string message() const;
};


class PETScTFQMR1Solver : public PETScSolver {
public:
  PETScTFQMR1Solver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc)
    : PETScSolver::PETScSolver(max_iterations, tolerance, arlb, arub, pc)
  {}

  virtual void setsolver(KSP &) const;
  virtual std::string message() const;
};


class PETScTFQMR2Solver : public PETScSolver {
public:
  PETScTFQMR2Solver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc)
    : PETScSolver::PETScSolver(max_iterations, tolerance, arlb, arub, pc)
  {}

  virtual void setsolver(KSP &) const;
  virtual std::string message() const;
};

class PETScCRSolver : public PETScSolver {
public:
  PETScCRSolver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc)
    : PETScSolver::PETScSolver(max_iterations, tolerance, arlb, arub, pc)
  {}

  virtual void setsolver(KSP &) const;
  virtual std::string message() const;
};

class PETScLeastSquaresSolver : public PETScSolver {
public:
  PETScLeastSquaresSolver(int max_iterations, double tolerance, double arlb, double arub, PreconditionerWrap * pc)
    : PETScSolver::PETScSolver(max_iterations, tolerance, arlb, arub, pc)
  {}

  virtual void setsolver(KSP &) const;
  virtual std::string message() const;
};



class PETScSolverMessager : public SolverMessager {
public:
  PETScSolverMessager(const int,  const int, const double);
  PETScSolverMessager() {};
  virtual std::string _message() const {return message;};
  ~PETScSolverMessager() {};
};

#endif // PETSCSOLVER_H


#endif //HAVE_PETSC




