# -*- python -*-
# $RCSfile: setup.py,v $
# $Revision: 1.26 $
# $Author: langer $
# $Date: 2005/06/27 17:30:45 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

## Usage:

#  In the top oof2 directory (the one containing this file) type
#  something like this:
#    python setup.py install     # installs in the default location
#    python setup.py install --prefix=/some/other/place
#    python setup.py [build_ext [--debug]] install --prefix ...
#  The flags --enable-mpi, --enable-petsc, and --enable-devel can
#  occur anywhere after 'setup.py' in the command line.


# External utilities
SWIG = 'swig'                           # must be swig 1.1 build 883
MAKEDEPEND = 'makedepend'

# The following determine some secondary installation directories.
# They will be created within the main installation directory
# specified by --prefix.

DATADIR = "share/oof2"
EXAMPLEDIR = "share/oof2/examples"
DOCDIR = "share/oof2/doc"

###############################

import distutils.core
from distutils.command import build_ext
from distutils.command import install
from distutils.command import clean
from distutils.command import build
from distutils import errors
from distutils.dir_util import remove_tree
import os
import stat
import string
import sys
import tempfile

try:
    import py2app                       # for creating a frozen Mac app.
except ImportError:
    have_py2app = 0
else:
    have_py2app = 1

DIRFILE = "DIR.py"                      # oof subdirectory manifest files
SWIGDIR = "SWIG"                        # root dir for swig output, inside SRC
SWIGCFILEEXT = 'cmodule.C'              # suffix for swig output files
SWIGARGS = "-shadow -dnone -python -c++ -c"

swigroot = os.path.join('SRC', SWIGDIR)
exampledir = os.path.join(*EXAMPLEDIR.split('/'))
datadir = os.path.join(*DATADIR.split('/'))
docdir = os.path.join(*DOCDIR.split('/'))

# The --enable-xxxx flags in the command line have to be obtained
# *before* we parse the DIR.py files, and the DIR.py files must be
# parsed before calling distutils.core.setup (because the list of
# source files comes from DIR.py).  But distutils.core.setup handles
# the command line arguments, so we have to look for and remove the
# --enable-xxxx flags here.

def get_oof_arg(arg):
    if arg in sys.argv:
        sys.argv.remove(arg)
        return 1
    return 0

HAVE_MPI = get_oof_arg('--enable-mpi')
HAVE_PETSC = get_oof_arg('--enable-petsc')
DEVEL = get_oof_arg('--enable-devel')

#########

# find_dirs() walks through the oof2 source directory looking for
# DIR.py files, reads the files, and fills in the dirdata lists.  The
# lists categorize all of the source files.  DIR.py files contain
# lists of filenames.  The lists are stored in variables whose names
# are the same as the keys in the dirdata dictionary.

dirdata = {'cfiles': [],                # *.[Cc] -- c and c++ source code
           'hfiles': [],                # *.h    -- c and c++ header files
           'swigfiles': [],             # *.swg  -- swig source code
           'swigpyfiles': [],           # *_.py  -- python included in swig
           'swigdirs': [],              # directories containing swig files
           'pyfiles': []}               # *.py   -- non-swigged python

def find_dirs(srcdir):
    dirfile = os.path.join(srcdir, DIRFILE)
    if os.path.exists(dirfile):
        # dirfile defines variables whose names are the same as the
        # dirdata keys.  The variables contain lists of file names.
        localdict = {}
        execfile(dirfile, globals(), localdict)
        # Now the variables are in localdict
        for key in dirdata.keys():
            try:
                value = localdict[key]
            except KeyError:
                pass
            else:
                for filename in value:
                    dirdata[key].append(os.path.join(srcdir, filename))
        # dirfile also contains a list of subdirectories to process.
        try:
            subdirs = localdict['subdirs']
        except KeyError:
            pass
        else:
            for subdir in subdirs:
                find_dirs(os.path.join(srcdir, subdir))

##########

# Find all python packages and subpackages in a directory by looking
# for __init__.py files.

def find_pkgs():
    pkglist = []
    os.path.walk('SRC', _find_pkgs, pkglist)
    return pkglist

def _find_pkgs(pkglist, dirname, subdirs):
    if os.path.exists(os.path.join(dirname, '__init__.py')):
        pkglist.append(dirname)

#####

# Find all directories containing at least one swig input file.  These
# are used to create the swigged python packages.  This is done by
# traversing the DIR.py files, so that random leftover .swg files in
# strange places don't create packages, and so that modules can be
# included conditionally by HAVE_XXX tests in DIR.py files.

def find_swig_pkgs():
    dict = {}
    for swigfile in dirdata['swigfiles']:
        dict[os.path.split(swigfile)[0]] = 1
    return dict.keys()

##########

def run_swig(dry_run):
    for swigfile in dirdata['swigfiles']:
        srcdir, filename = os.path.split(swigfile)
        # swig output for ./SRC/a/b/c goes to SRC/SWIG/a/b/c
        swigdir = os.path.join(swigroot, srcdir[6:]) 
        basename = os.path.splitext(filename)[0]
        cfile = os.path.join(swigdir, basename+SWIGCFILEEXT)
        pyfile = os.path.join(swigdir, basename+'.py')
        # We've already removed files that are out of date, so if
        # cfile and pyfile exist, they don't have to be recreated.
        if not (os.path.exists(cfile) and os.path.exists(pyfile)):
            # The following command makes the directory for the swig
            # output, cds to the source directory, and runs swig.  It
            # uses absolute path names for all of the directories so
            # that we don't have to worry about the effect of the cd.
            # The cd is necessary because the swig 'pragma(python)
            # include' lines don't seem to respect the -I flags.
            cmd = 'mkdir -p %(dir)s; cd %(srcdir)s; %(swig)s %(swigargs)s -I%(swigroot)s -I%(swigdir)s -I%(srcroot)s -I%(srcdir)s -module %(module)s -o %(cfile)s %(swigfile)s' \
                  %  {'dir' : os.path.abspath(swigdir),
                      'srcdir' : os.path.abspath(srcdir),
                      'srcroot' : os.path.abspath('SRC'),
                      'swig': SWIG,
                      'swigroot' : os.path.abspath(swigroot),
                      'swigdir': os.path.abspath(swigdir),
                      'swigargs' : SWIGARGS,
                      'module' : basename,
                      'cfile' : os.path.abspath(cfile),
                      'swigfile' : os.path.abspath(swigfile)}
            print cmd
            if not dry_run:
                os.system(cmd)

##########

# Get a file's modification time.  The time is returned as an integer.
# All we care about is that the integers are properly ordered.

def modification_time(file):
    return os.stat(file)[stat.ST_MTIME]

#########

# Define a subclass of the distutils build_ext class for building
# python extensions.

class oof_build_ext(build_ext.build_ext):
    # build_extensions is called by build_ext.run().
    def build_extensions(self):
        self.compiler.add_include_dir(os.path.join(self.build_temp, 'SRC'))
        self.compiler.add_include_dir('SRC')
        if self.debug:
            self.compiler.define_macro('DEBUG')
            self.compiler.undefine_macro('NDEBUG')
        # Make the automatically generated .h files.
        self.make_oofconfig()
        self.make_oofcpph()
        # Run makedepend
        depdict = self.find_dependencies()
        # Remove out-of-date files
        self.clean_targets(depdict)
        # Generate swigged .C and .py files
        run_swig(self.dry_run)
        # Actually build the extension by calling the distutils base class func.
        build_ext.build_ext.build_extensions(self)

    def make_oofconfig(self):
        cfgfilename = os.path.normpath(os.path.join(self.build_temp,
                                                    'SRC/oofconfig.h'))
        # If oofconfig.h already exists, don't recreate it.
        if not os.path.exists(cfgfilename):
            print "creating", cfgfilename
            if not self.dry_run:
                os.system('mkdir -p %s' % os.path.join(self.build_temp, 'SRC'))
                cfgfile = open(cfgfilename, "w")
                print >> cfgfile, """\
                // This file was created automatically by setup.py.  Do not edit it.
                // Re-run setup.py to change the options.
                """
                if HAVE_PETSC:
                    print >> cfgfile, '#define HAVE_PETSC 1'
                if HAVE_MPI:
                    print >> cfgfile, '#define HAVE_MPI 1'
                if DEVEL:
                    print >> cfgfile, '#define DEVEL 1'
                if self.check_header('<sstream>'):
                    print >> cfgfile, '#define HAVE_SSTREAM'
                else:
                    print >> cfgfile, '// #define HAVE_SSTREAM'
                print >> cfgfile, '#include <Python.h>'
                cfgfile.close()

    def make_oofcpph(self):
        # Make <build_temp>/SRC/oofcpp.h, which is included by
        # SRC/common/oofcpp.C and initializes all of the imported
        # swigged modules.
        oofcpph = os.path.join(self.build_temp, 'SRC', 'oofcpp.h')
        print "creating", oofcpph
        if not self.dry_run:
            if os.path.exists(oofcpph):
                os.remove(oofcpph)
            hfile = open(oofcpph, 'w')
            print >> hfile, "// automatically generated file, do not edit!"
            print >> hfile, "#ifndef OOFCPP_H"
            print >> hfile, "#define OOFCPP_H"
            for file in dirdata['swigfiles']:
                modulename = os.path.splitext(os.path.basename(file))[0]
                print >> hfile, \
                      "extern void init%(module)sc(); init%(module)sc();" \
                      % {'module':modulename}
            print >> hfile, "#endif // OOFCPP_H"
            hfile.close()

    def check_header(self, headername):
        global platform_compiler_args
        global prelink_suppression_arg
        # Check that a c++ header file exists on the system.
        print "Testing for", headername
        # mktemp is insecure, but the better routines are only in
        # python2.3, and we don't otherwise require 2.3.
        tmpfilename = tempfile.mktemp(suffix='.C')
        tmpfile = open(tmpfilename, 'w')
        print >> tmpfile, """\
        #include %s
        int main(int, char**) { return 1; }
        """ % headername
        tmpfile.flush()
        try:
            try:
                ofiles = self.compiler.compile([tmpfile.name],
                                               extra_postargs=platform_compiler_args + prelink_suppression_arg)
            except errors.CompileError:
                return 0
            ofile = ofiles[0]
            dir = os.path.split(ofile)[0]
            os.remove(ofiles[0])
            if dir:
                try:
                    os.removedirs(dir)
                except:
                    pass
            return 1
        finally:
            os.remove(tmpfilename)

    def find_dependencies(self):
        # distutils doesn't provide a makedepend-like facility, so we
        # have to do it ourselves.  This routine runs make_depend and
        # constructs a dictionary listing the dependencies of each .o
        # file and each swig-generated .C file.

        # TODO: This function is a mess.  Clean it up.
        
        # depdict[file] is a list of sources that the file depends
        # upon.
        depdict = {}
        # run makedepend on the C++ files
        depfilename = tempfile.mktemp()
        cmd = '%(makedepend)s -f- -ISRC -I%(builddir)s -I%(buildsrc)s -- %(files)s > %(dependencies)s 2> /dev/null' \
              % {'makedepend' : MAKEDEPEND,
                 'files' : string.join(dirdata['cfiles'], ' '),
                 'dependencies' : depfilename,
                 'builddir' : self.build_temp,
                 'buildsrc' : os.path.join(self.build_temp, 'SRC')
                 }
##        print cmd
        print "running makedepend on .C files."
        os.system(cmd)
        # Read and parse the makedepend output.  Each nontrivial line
        # is of the form "target: source [source [...]]"
        depfile = open(depfilename, 'r')
        for line in depfile.readlines():
            sline = line.strip()
            if sline and sline[0] != '#':
                target, sources = sline.split(': ')
                # The actual target that will be built is in the
                # build_temp directory.  makedepend assumes that it's
                # in the same directory as the source file, so we have
                # to modify the target name.
                realtarget = os.path.normpath(os.path.join(self.build_temp,
                                                           target))
                for source in sources.split():
                    try:
                        depdict[realtarget].append(source)
                    except KeyError:
                        depdict[realtarget] = [source]
        os.remove(depfilename)
        # makedepend omits the obvious dependency ("target.o:
        # target.C"), but we need it.
        for file in dirdata['cfiles']:
            ofile = os.path.normpath(
                os.path.join(self.build_temp, os.path.splitext(file)[0]+'.o'))
            try:
                depdict[ofile].append(file)
            except KeyError:
                depdict[ofile] = [file]

        # .C and.py Files in the SWIG directory depend on those in the
        # SRC directory.  Run makedepend on the swig source files.
        depfilename = tempfile.mktemp()
        cmd = '%(makedepend)s -f- -I. -ISRC -I%(builddir)s -- %(files)s > %(dependencies)s 2> /dev/null' \
              % {'makedepend' : MAKEDEPEND,
                 'files' : string.join(dirdata['swigfiles'], ' '),
                 'dependencies' : depfilename,
                 'builddir' : self.build_temp
                 }
        print "running makedepend on .swg files."
##        print cmd
        os.system(cmd)
        depfile = open(depfilename, 'r')
        for line in depfile.readlines():
            sline = line.strip()
            if sline and sline[0] != '#':
                target, sources = sline.split(': ')
                targetbase = os.path.splitext(target)[0]
                # Use [6:] to strip "./SRC/" from source paths
                targetc = os.path.normpath(
                    os.path.join(swigroot, targetbase[6:] + SWIGCFILEEXT))
                targetpy = os.path.normpath(
                    os.path.join(swigroot, targetbase[6:] + '.py'))
                for source in sources.split():
                    try:
                        depdict[targetc].append(source)
                    except KeyError:
                        depdict[targetc] = [source]
                    try:
                        depdict[targetpy].append(source)
                    except KeyError:
                        depdict[targetpy] = [source]
        os.remove(depfilename)
        # Add in the implicit dependencies on the .swg files.
        for file in dirdata['swigfiles']:
            # file is ./SRC/dir/whatver.swg
            base = os.path.splitext(file)[0][6:] # dir/whatever
            cfile = os.path.normpath(os.path.join(swigroot, base+SWIGCFILEEXT))
            pyfile = os.path.normpath(os.path.join(swigroot, base+'.py'))
            try:
                depdict[cfile].append(file)
            except KeyError:
                depdict[cfile] = [file]
            try:
                depdict[pyfile].append(file)
            except:
                depdict[pyfile] = [file]
        # Add in the implicit dependencies on the _.py files.
        for underpyfile in dirdata['swigpyfiles']:
            base = os.path.splitext(underpyfile)[0][:-1] # drop _.py
            pyfile = os.path.normpath(os.path.join(swigroot, base[6:]+'.py'))
            try:
                depdict[pyfile].append(underpyfile)
            except KeyError:
                depdict[pyfile] = [underpyfile]

        return depdict

    # Remove out-of-date target files
    def clean_targets(self, depdict):
        outofdate = 0
        if not self.dry_run:
            for target, sources in depdict.items():
                if os.path.exists(target):
                    targettime = modification_time(target)
                    sourcetime = max([modification_time(x) for x in sources])
                    if sourcetime > targettime:
##                        print 'removing', target
                        os.remove(target)
                        outofdate = 1
                else:
                    outofdate = 1
        if outofdate:
            ## TODO: Remove the .so file.  I can't figure out how to
            ## find its name at this point, though.
            pass

    
###################################################

class oof_clean(clean.clean):
    def run(self):
        if os.path.exists(swigroot):
            remove_tree(swigroot, dry_run=self.dry_run)
        clean.clean.run(self)
    
###################################################

# Change the order in which the build sub-commands are run, so that
# the swig-generated .py files are built *before* they're installed.

class oof_build(build.build):
    sub_commands = [('build_ext', build.build.has_ext_modules),
                    ('build_py', build.build.has_pure_modules),
                    ('build_clib', build.build.has_c_libraries),
                    ('build_scripts', build.build.has_scripts)]
    
###################################################

##class oof_install(install.install):
##    def run(self):

##        # Create the oofpaths.py file.  First make sure that the
##        # target directory exists.
##        oof2dir = os.path.join(self.install_lib, 'oof2')
##        if not os.path.exists(oof2dir):
##            os.makedirs(oof2dir)
        
##        cmd = 'sed -e s?OOF2LIB?%(libdir)s? -e s?OOF2BIN?%(binary)s? -e s?OOF2EXAMPLES?%(examples)s? -e s?OOF2DATA?%(data)s? -e s?OOF2DOC?%(doc)s? < oofpaths.in > %(oofpathspy)s' % {
##            'libdir' : os.path.join(self.install_lib, 'oof2'),
##            'oofpathspy' : os.path.join(self.install_lib,'oof2','oofpaths.py'),
##            'binary' : os.path.join(self.install_lib, 'oofcppcmodule.so'),
##            'examples' : os.path.join(self.install_base, exampledir),
##            'data' : os.path.join(self.install_base, datadir),
##            'doc' : os.path.join(self.install_base, docdir)
##            }
##        os.system(cmd)
        
##        install.install.run(self)
        
###################################################

# Build the C++ extension module.

find_dirs('.')

# The [6:] in the following strips the "./SRC/" from the beginning of
# the file names.  splitext(file)[0] is the path to a file with the
# extension stripped.

swiggedcfiles = [os.path.join(swigroot,
                              os.path.splitext(file)[0][6:]+SWIGCFILEEXT)
                 for file in dirdata['swigfiles']]

extra_compile_args = []
macros = []
libs = []
libdirs = []
incdirs = []
extra_link_args = []

# Utilities for running xxxxx-config programs that get the flags and
# libs required to link with external libraries.

def add_third_party_includes(cmd):
    f = os.popen(cmd, 'r')
    for line in f.readlines():
        for flag in line.split():
            if flag[:2] == '-I':
                incdirs.append(flag[2:])

def add_third_party_libs(cmd):
    f = os.popen(cmd, 'r')
    for line in f.readlines():
        for flag in line.split():
            if flag[:2] == '-l':
                libs.append(flag[2:])
            if flag[:2] == '-L':
                libdirs.append(flag[2:])

pathdirs = os.getenv("PATH").split(":")
def check_exec(xfile):          # is xfile an executable file in the path?
    for dir in pathdirs:
        if os.path.exists(dir):
            files = os.listdir(dir)
            if xfile in files:
                fullname = os.path.join(dir, xfile)
                if os.access(fullname, os.X_OK):
                    return 1
    print "Warning!", xfile, "not found!"
    

if os.path.exists('/usr/local/lib'):
    libdirs.append('/usr/local/lib')
if os.path.exists('/usr/site/lib'):
    libdirs.append('/usr/site/lib')
if os.path.exists('/usr/site/include'):
    incdirs.append('/usr/site/include')

if check_exec('gtk-config'):
    add_third_party_includes('gtk-config --cflags')
    add_third_party_libs('gtk-config --libs')
if check_exec('imlib-config'):
    add_third_party_includes('imlib-config --cflags')
    add_third_party_libs('imlib-config --libs-gdk')
if check_exec('Magick++-config'):
    add_third_party_libs('Magick++-config --ldflags --libs')

libs.extend(['gtk-canvas', 'art_lgpl'])

platform_compiler_args = []

# The prelink-suppression argument is used when running the compiler
# to test for libraries.  Such builds need to be reasonably clean in
# terms of not creating a lot of auxiliary files, and it's OK if
# they're a bit slow.  Currently this is only set on SGIs, and it
# prevents the creation of the "ii_files" subdirectory for the
# library-check compilations.

prelink_suppression_arg = []


if sys.platform == 'darwin':
    extra_link_args.extend(['-faltivec', '-framework', 'vecLib'])
    # For some reason, distutils thinks it's a good idea to use the
    # -Wall option to gcc.  This generates a ridiculous number of
    # irrelevant warnings.  -w turns off warnings altogether.
    platform_compiler_args.append('-w')
elif sys.platform == 'linux2':
    libs.extend(['lapack', 'blas', 'g2c', 'm'])
    platform_compiler_args.append('-w')     # suppress warning messages
elif sys.platform[:4] == 'irix':
    platform_compiler_args.append('-LANG:std')
    prelink_suppression_arg.append('-no_prelink')
    libs.extend(['lapack', 'blas', 'ftn', 'm'])


oofcmodule = distutils.core.Extension(
    name='oofcppcmodule',
    language='c++',
    sources=dirdata['cfiles'] + swiggedcfiles,
    define_macros = macros,
    extra_compile_args = platform_compiler_args,
    include_dirs = incdirs,
    libraries = libs,
    library_dirs = libdirs,
    extra_link_args = extra_link_args
    )

############

# Construct the list of packages

# find non-swigged files
pkg_list = ['oof2']
pkgs = find_pkgs()                      # ['SRC', 'SRC/common', ...]
for pkg in pkgs:
    if pkg != 'SRC':
        pkgname = 'oof2.' + pkg[4:].replace('/', '.')
        pkg_list.append(pkgname)

# Find swig-generated package components
pkgs = find_swig_pkgs()                 #  ['./SRC/common/IO/GUI', ... ]

pkg_list.append('oof2.SWIG')
for pkg in pkgs:
    pkgdir = os.path.join(swigroot, pkg[6:]) # eg, SRC/SWIG/common
    pkgname = 'oof2.' + pkgdir[4:].replace('/', '.') # eg, oof2.SWIG.common
    pkg_list.append(pkgname)
    if not os.path.exists(pkgdir):
        os.makedirs(pkgdir)

#############
        
# Make sure that there's an __init__.py file in each SWIG subdirectory
# and that the subdirectory is in the package list.  This is necessary
# because we've found the swig packages by searching for .swg files,
# but they might be in subdirectories of swig-less directories.  If
# the swig-less directories aren't listed as packages, then their
# __init__.py files won't be installed, and the swig directories can't
# be imported.

def _make_init_py(dummy, dirname, subdirs):
    initpy = os.path.join(dirname, '__init__.py')
    if not os.path.exists(initpy):
        file = open(initpy, 'w')
        print >> file, "# This file intentionally left blank."
        file.close()
    pkgname = 'oof2.' + dirname[4:].replace('/', '.') # remove 'SRC/'
    if pkgname not in pkg_list:
        pkg_list.append(pkgname)

os.path.walk(swigroot, _make_init_py, None)

############

# Find example files:

examples = [os.path.join('examples', file) for file in os.listdir('examples')]
examples = [file for file in examples if os.path.isfile(file)] # exclude dirs
exampledir = os.path.join('share', 'oof2', 'examples') # destination

############

py2app_options = dict(
    argv_emulation=True
    )
sys.path.append('SRC')                  # so that py2app can find imports

try:                                    # modify path for gtk 1.2, if necessary
    import pygtk
    pygtk.require("1.2")
except:
    pass

############

setupargs = dict(
    name = "oof2",
    version = "2.0.beta5.1",
    description = "Analysis of material microstructures, from NIST.",
    author = 'The NIST OOF Team',
    author_email = 'oof_manager@ctcms.nist.gov',
    url = "http://www.ctcms.nist.gov/oof/oof2/",
    scripts = ["oof2"],
    cmdclass = {"build_ext" : oof_build_ext,
                "build" : oof_build,
                "clean" : oof_clean},
    packages = pkg_list,
    package_dir = {'oof2':'SRC'},
    ext_modules=[oofcmodule],
    data_files = [(exampledir, examples)]
    )

if have_py2app:
    setupargs['app'] = ['SRC/common/oof.py']
    setupargs['options'] = {'py2app': py2app_options}

if __name__ == '__main__':
    distutils.core.setup(**setupargs)
